# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""API for the TextGrid aggregator service"""
import logging
from typing import List, Optional, Union, overload

import requests
from requests.models import Response

from tgclients.config import TextgridConfig

logger = logging.getLogger(__name__)


class Aggregator:
    """Provide access to the Textgrid Aggregator Service.
    API docs: https://textgridlab.org/doc/services/submodules/aggregator/docs/api.html"""

    def __init__(self, config: TextgridConfig = TextgridConfig()) -> None:
        self._url = config.aggregator
        self._config = config
        self._requests = requests.Session()

    @overload
    def zip(self, textgrid_uris: str, sid: Optional[str] = None) -> Response:
        ...

    @overload
    def zip(self, textgrid_uris: List[str], sid: Optional[str] = None) -> Response:
        ...

    def zip(self, textgrid_uris: Union[str, List[str]], sid: Optional[str] = None) -> Response:
        """Download aggregated TextGrid objects as ZIP file.
        https://textgridlab.org/doc/services/submodules/aggregator/docs/zip.html

        Args:
            textgrid_uris (List[str] or str): a single or a list of TextGrid URIs
            sid (Optional[str], optional): Session ID. Defaults to None.

        Returns:
            Response: the response with zip file in body.content
        """
        if isinstance(textgrid_uris, list):
            textgrid_uris = ','.join(textgrid_uris)
        url = self._url + '/zip/'
        response = self._requests.get(url + textgrid_uris, params={ 'sid': sid },
                                timeout=self._config.http_timeout)
        return response

    @overload
    def text(self, textgrid_uris: str, sid: Optional[str] = None) -> Response:
        ...

    @overload
    def text(self, textgrid_uris: List[str], sid: Optional[str] = None) -> Response:
        ...

    # python 3.10 allows writinh Union as |
    # https://www.blog.pythonlibrary.org/2021/09/11/python-3-10-simplifies-unions-in-type-annotations/
    def text(self, textgrid_uris: Union[str, List[str]], sid: Optional[str] = None) -> Response:
        """Download aggregated TextGrid objects as plain text file.

        Args:
            textgrid_uris (List[str] or str): a single or a list of TextGrid URIs
            sid (Optional[str], optional): Session ID. Defaults to None.

        Returns:
            Response: the respone with the text in the body
        """
        if isinstance(textgrid_uris, list):
            textgrid_uris = ','.join(textgrid_uris)
        url = self._url + '/text/'
        response = self._requests.get(url + textgrid_uris, params={ 'sid': sid },
                                timeout=self._config.http_timeout)
        return response

    @overload
    def teicorpus(self, textgrid_uris: str, sid: Optional[str] = None) -> Response:
        ...

    @overload
    def teicorpus(self, textgrid_uris: List[str], sid: Optional[str] = None) -> Response:
        ...

    def teicorpus(self, textgrid_uris: Union[str, List[str]], sid: Optional[str] = None
        ) -> Response:
        """Download aggregated TextGrid objects as TEI corpus.

        Args:
            textgrid_uris (List[str] or str): a single or a list of TextGrid URIs
            sid (Optional[str], optional): Session ID. Defaults to None.

        Returns:
            Response: the respone with the TEI corpus in the body
        """
        if isinstance(textgrid_uris, list):
            textgrid_uris = ','.join(textgrid_uris)
        url = self._url + '/teicorpus/'
        response = self._requests.get(url + textgrid_uris, params={ 'sid': sid },
                                timeout=self._config.http_timeout)
        return response

    @overload
    def render(self,
        textgrid_uris: str,
        sid: Optional[str] = None,
        stylesheet_uri: Optional[str] = None,
        mediatype: Optional[str] = None,
        link_pattern: Optional[str] = None,
        sandbox: Optional[bool] = None
        ) -> Response:
        ...

    @overload
    def render(self, textgrid_uris: List[str],
        sid: Optional[str] = None,
        stylesheet_uri: Optional[str] = None,
        mediatype: Optional[str] = None,
        link_pattern: Optional[str] = None,
        sandbox: Optional[bool] = None
        ) -> Response:
        ...

    def render(self,
        textgrid_uris: Union[str, List[str]],
        sid: Optional[str] = None,
        stylesheet_uri: Optional[str] = None,
        mediatype: Optional[str] = None,
        link_pattern: Optional[str] = None,
        sandbox: Optional[bool] = None
        ) -> Response:
        """Apply an XSLT stylesheet to one or more TextGrid URIs.
        Will render (X)HTML by default with XSLT stylesheets from tei-c.org
        see https://textgridlab.org/doc/services/submodules/aggregator/docs/html.html

        Args:
            textgrid_uris (List[str] or str): a single or a list of TextGrid URIs
            sid (Optional[str], optional): Session ID. Defaults to None.
            stylesheet_uri (Optional[str], optional): alternative XSLT stylesheet to use.
                                                      Must be a TextGrid URI.
            mediatype (Optional[str], optional): The requested content type.
                                                 E.g., text/html or text/xml. Default is text/html
            link_pattern (Optional[str], optional): URL pattern for links.
                                                    @URI@ will be replaced with the textgrid: URI.
            sandbox (Optional[bool], optional): access sandboxed data. Defaults to false

        Returns:
            Response: the respone with the TEI corpus in the body
        """
        if isinstance(textgrid_uris, list):
            textgrid_uris = ','.join(textgrid_uris)
        url = self._url + '/html/'
        response = self._requests.get(url + textgrid_uris,
                                params={
                                    'sid': sid,
                                    'stylesheet': stylesheet_uri,
                                    'mediatype': mediatype,
                                    'linkPattern': link_pattern,
                                    'sandbox': sandbox
                                },
                                timeout=self._config.http_timeout)
        return response



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: CC0-1.0

__version__ = '0.15.1'

from tgclients.aggregator import (
    Aggregator,
)
from tgclients.auth import (
    TextgridAuth,
    TextgridAuthException,
)
from tgclients.config import (
    TextgridConfig,
)
from tgclients.crud import (
    TextgridCrud,
    TextgridCrudRequest,
    TextgridCrudException,
)
from tgclients.metadata import (
    TextgridMetadata,
)
from tgclients.search import (
    TextgridSearch,
    TextgridSearchRequest,
    TextgridSearchException,
)
from tgclients.utils import (
    Utils,
)
__all__ = [
    'Aggregator',
    'TextgridAuth',
    'TextgridAuthException',
    'TextgridConfig',
    'TextgridCrud',
    'TextgridCrudRequest',
    'TextgridCrudException',
    'TextgridSearch',
    'TextgridSearchRequest',
    'TextgridSearchException',
    'Utils',
]



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Utility functions for working with the TextGrid repository"""
from pathlib import Path
from typing import List

import defusedxml.ElementTree as ET
from jinja2 import Environment, FileSystemLoader

from tgclients.databinding.tgsearch import Response as SearchResponse


class Utils:
    """Utility functions for working with the TextGrid repository"""

    @staticmethod
    def list_to_aggregation(textgrid_uri: str, members: List[str]) -> str:
        """Create XML for a TextGrid aggregation from list

        Args:
            textgrid_uri (str): textgrid URI of the aggregation to create
            members (list[str]): list of textgrid URIs inside aggregation

        Returns:
            str: XML for TextGrid Aggregation
        """
        path = Path(__file__).parent / 'templates'
        env = Environment(
            loader=FileSystemLoader(Path(path)), autoescape=True)
        template = env.get_template('aggregation.xml.jinja2')
        aggregation = template.render(id=textgrid_uri, members=members)
        return aggregation

    @staticmethod
    def aggregation_to_list(xml: str) -> List[str]:
        """Extract URIs from TextGrid aggregation into a list

        Args:
            xml (str): TextGrid aggregation XML

        Returns:
            list[str]: TextGrid URIs from aggregation
        """
        res = []
        root = ET.fromstring(xml)
        tag_name = '{http://www.openarchives.org/ore/terms/}aggregates'
        attr_name = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}resource'
        for descendant in root.iter(tag_name):
            res.append(descendant.attrib[attr_name])
        return res



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: CC0-1.0

from tgclients.databinding.rdf import (
    Rdf,
    RdfType,
)
from tgclients.databinding.textgrid_metadata_2010 import (
    AuthorityType,
    BibliographicCitationType,
    CollectionType,
    DateType,
    EditionType,
    GeneratedType,
    GeneratedTypeAvailability,
    GenericType,
    IdentifierType,
    ItemType,
    MetadataContainerType,
    Object,
    ObjectCitationType,
    PidPidType,
    ProvidedType,
    RelationType as TextgridMetadata2010RelationType,
    SourceType,
    TgObjectMetadata,
    WorkType,
    WorkTypeGenre,
)
from tgclients.databinding.textgrid_metadata_agent_2010 import (
    AgentRoleType,
    AgentType,
    PersonType,
)
from tgclients.databinding.textgrid_metadata_script_2010 import FormOfNotationType
from tgclients.databinding.tgsearch import (
    FulltextType,
    EntryType,
    FacetGroupType,
    FacetResponse,
    FacetResponseType,
    FacetType,
    PathGroupType,
    PathResponse,
    PathType,
    RelationResponseType,
    RelationType as TgsearchRelationType,
    Response,
    ResultType,
    Revisions,
    TextgridUris,
)

__all__ = [
    "Rdf",
    "RdfType",
    "AuthorityType",
    "BibliographicCitationType",
    "CollectionType",
    "DateType",
    "EditionType",
    "GeneratedType",
    "GeneratedTypeAvailability",
    "GenericType",
    "IdentifierType",
    "ItemType",
    "MetadataContainerType",
    "Object",
    "ObjectCitationType",
    "PidPidType",
    "ProvidedType",
    "TextgridMetadata2010RelationType",
    "SourceType",
    "TgObjectMetadata",
    "WorkType",
    "WorkTypeGenre",
    "AgentRoleType",
    "AgentType",
    "PersonType",
    "FormOfNotationType",
    "FulltextType",
    "EntryType",
    "FacetGroupType",
    "FacetResponse",
    "FacetResponseType",
    "FacetType",
    "PathGroupType",
    "PathResponse",
    "PathType",
    "RelationResponseType",
    "TgsearchRelationType",
    "Response",
    "ResultType",
    "Revisions",
    "TextgridUris",
]



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: CC0-1.0

from dataclasses import dataclass, field
from typing import List, Optional
from tgclients.databinding.textgrid_metadata_2010 import Object

__NAMESPACE__ = "http://www.textgrid.info/namespaces/middleware/tgsearch"


@dataclass
class FulltextType:
    kwic: List["FulltextType.Kwic"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )
    hits: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )

    @dataclass
    class Kwic:
        left: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
                "required": True,
            }
        )
        match: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
                "required": True,
            }
        )
        right: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
                "required": True,
            }
        )
        xpath: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            }
        )
        ref_id: Optional[str] = field(
            default=None,
            metadata={
                "name": "refId",
                "type": "Attribute",
            }
        )
        ref_title: Optional[str] = field(
            default=None,
            metadata={
                "name": "refTitle",
                "type": "Attribute",
            }
        )


@dataclass
class EntryType:
    class Meta:
        name = "entryType"

    textgrid_uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "textgridUri",
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )
    format: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )
    title: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )


@dataclass
class FacetType:
    class Meta:
        name = "facetType"

    value: str = field(
        default="",
        metadata={
            "required": True,
        }
    )
    count: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class RelationType:
    class Meta:
        name = "relationType"

    s: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )
    p: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )
    o: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "required": True,
        }
    )


@dataclass
class Revisions:
    class Meta:
        name = "revisions"
        namespace = "http://www.textgrid.info/namespaces/middleware/tgsearch"

    revision: List[int] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        }
    )
    textgrid_uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "textgridUri",
            "type": "Attribute",
        }
    )


@dataclass
class TextgridUris:
    class Meta:
        name = "textgridUris"
        namespace = "http://www.textgrid.info/namespaces/middleware/tgsearch"

    textgrid_uri: List[str] = field(
        default_factory=list,
        metadata={
            "name": "textgridUri",
            "type": "Element",
        }
    )
    hits: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    start: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    limit: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class FacetGroupType:
    class Meta:
        name = "facetGroupType"

    facet: List[FacetType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class PathType:
    class Meta:
        name = "pathType"

    entry: List[EntryType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
            "min_occurs": 1,
        }
    )


@dataclass
class RelationResponseType:
    class Meta:
        name = "relationResponseType"

    relation: List[RelationType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )


@dataclass
class FacetResponse:
    class Meta:
        name = "facetResponse"
        namespace = "http://www.textgrid.info/namespaces/middleware/tgsearch"

    facet_group: List[FacetGroupType] = field(
        default_factory=list,
        metadata={
            "name": "facetGroup",
            "type": "Element",
        }
    )


@dataclass
class FacetResponseType:
    class Meta:
        name = "facetResponseType"

    facet_group: List[FacetGroupType] = field(
        default_factory=list,
        metadata={
            "name": "facetGroup",
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )


@dataclass
class PathGroupType:
    class Meta:
        name = "pathGroupType"

    path: List[PathType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )
    start_uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "startUri",
            "type": "Attribute",
        }
    )


@dataclass
class PathResponse:
    class Meta:
        name = "pathResponse"
        namespace = "http://www.textgrid.info/namespaces/middleware/tgsearch"

    path_group: List[PathGroupType] = field(
        default_factory=list,
        metadata={
            "name": "pathGroup",
            "type": "Element",
        }
    )


@dataclass
class ResultType:
    class Meta:
        name = "resultType"

    object_value: Optional[Object] = field(
        default=None,
        metadata={
            "name": "object",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    fulltext: List[FulltextType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )
    path_response: Optional[PathResponse] = field(
        default=None,
        metadata={
            "name": "pathResponse",
            "type": "Element",
            "namespace": "http://www.textgrid.info/namespaces/middleware/tgsearch",
        }
    )
    textgrid_uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "textgridUri",
            "type": "Attribute",
        }
    )
    authorized: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    sandbox: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    author: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class Response:
    class Meta:
        name = "response"
        namespace = "http://www.textgrid.info/namespaces/middleware/tgsearch"

    result: List[ResultType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        }
    )
    relation_response: Optional[RelationResponseType] = field(
        default=None,
        metadata={
            "name": "relationResponse",
            "type": "Element",
        }
    )
    facet_response: Optional[FacetResponseType] = field(
        default=None,
        metadata={
            "name": "facetResponse",
            "type": "Element",
        }
    )
    hits: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    session: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    start: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    limit: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: CC0-1.0

"""Variable config options with defaults to be used with the TextGrid clients library."""
import logging
from typing import Optional

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT: float = 120.00

PROD_SERVER: str = 'https://textgridlab.org'
DEV_SERVER: str = 'https://dev.textgridlab.org'
TEST_SERVER: str = 'https://test.textgridlab.org'


class TextgridConfig:
    """Provide standard configuration / URLs for TextGrid services.
    Default is to connect to the TextGrid
    production server (https://textgridlab.org).
    Pass the constants tgclients.config.DEV_SERVER or
    tgclients.config.TEST_SERVER to the constructor to change to
    develop or test server or provide an URL for your own instance.
    """

    def __init__(self, host: Optional[str] = 'https://textgridlab.org') -> None:
        """TextGrid Service Endpoint Configuration.

        Overwrite the host string to connect to other servers, e.g.:
        * Test-Server: https://test.textgridlab.org
            (use constant tgclients.config.TEST_SERVER)
        * Development-Server: https://dev.textgridlab.org
            (use constant tgclients.config.DEV_SERVER)

        Args:
            host (Optional[str]):TextGrid server. Defaults to 'https://textgridlab.org'.
        """
        if not isinstance(host, str) or host == '':
            logger.info(
                'host param was None or emtpy, default to: %s', PROD_SERVER)
            host = PROD_SERVER
        if host.endswith('/'):
            logger.info('trailing slash in hostname detected and removed')
            host = host[:-1]
        self._host = host
        self._http_timeout = DEFAULT_TIMEOUT

    @property
    def host(self) -> str:
        """the host URL

        Returns:
            str: the configured host URL
        """
        return self._host

    @property
    def auth_wsdl(self) -> str:
        """the tgauth wsdl location

        Returns:
            str: the tgauth wsdl location
        """
        return self._host + '/1.0/tgauth/wsdl/tgextra.wsdl'

    @property
    def auth_address(self) -> str:
        """the tgauth service location

        Returns:
            str: the tgauth service location
        """
        return self._host + '/1.0/tgauth/tgextra.php'

    @property
    def extra_crud_wsdl(self) -> str:
        """the tgextra wsdl location

        Returns:
            str: the tgextra wsdl location
        """
        return self._host + '/1.0/tgauth/wsdl/tgextra-crud.wsdl'

    @property
    def extra_crud_address(self) -> str:
        """the tgextra service location

        Returns:
            str: the tgextra service location
        """
        return self._host + '/1.0/tgauth/tgextra-crud.php'

    @property
    def search(self) -> str:
        """the nonpublic tgsearch service location

        Returns:
            str: the nonpublic tgsearch service location
        """
        return self._host + '/1.0/tgsearch'

    @property
    def search_public(self) -> str:
        """the public tgsearch service location

        Returns:
            str: the public tgsearch service location
        """
        return self._host + '/1.0/tgsearch-public'

    @property
    def crud(self) -> str:
        """the nonpublic tgcrud REST service location

        Returns:
            str: the nonpublic tgcrud REST service location
        """
        return self._host + '/1.0/tgcrud/rest'

    @property
    def crud_public(self) -> str:
        """the public tgcrud REST service location

        Returns:
            str: the public tgcrud REST service location
        """
        return self._host + '/1.0/tgcrud-public/rest'

    @property
    def aggregator(self) -> str:
        """the aggregator service location

        Returns:
            str: the aggregator service location
        """
        return self._host + '/1.0/aggregator'

    @property
    def http_timeout(self) -> float:
        """HTTP timeout to be used when accessing TextGrid services

        Returns:
            float: http timeout in seconds
        """
        return self._http_timeout

    @http_timeout.setter
    def http_timeout(self, value: float) -> None:
        """Set HTTP timeout to be used when accessing TextGrid services

        Args:
            value (float): the timeout
        """
        self._http_timeout = value



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""TextGrid Search API."""
import logging
from io import BytesIO
from typing import Optional, List

import requests
from requests.models import Response
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser

from tgclients.config import TextgridConfig
from tgclients.databinding.tgsearch import Response as SearchResponse, TextgridUris

logger = logging.getLogger(__name__)


class TextgridSearchRequest:
    """Provide low level access to the TextGrid search service, returning the response objects."""

    def __init__(self, config: TextgridConfig = TextgridConfig(), nonpublic: bool = False) -> None:
        if nonpublic:
            self._url = config.search
        else:
            self._url = config.search_public
        self._config = config
        # reuse tcp connections: https://requests.readthedocs.io/en/latest/user/advanced/#session-objects
        self._requests = requests.Session()

    def info(self, textgrid_uri: str, sid: Optional[str] = None) -> Response:
        """Retrieve metadata for a textgrid object specified by its
        textgrid-uri

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str]): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: metadata for uri
        """
        url = self._url + '/info/'
        response = self._requests.get(
            url + textgrid_uri, params={'sid': sid}, timeout=self._config.http_timeout)
        return self._handle_response(response)

    def list_project_root(self, project_id: str, sid: Optional[str] = None) -> Response:
        """Get objects belonging to a project, filtered by objects that are in
        an aggregation in the same project.

        Args:
            project_id (str): the ID of the project to list
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: HTTP response from service, containing a list of textgrid metadata entries
        """
        response = self._requests.get(
            self._url + '/navigation/' + project_id, params={'sid': sid},
            timeout=self._config.http_timeout)
        return self._handle_response(response)

    def list_aggregation(self, textgrid_uri: str, sid: Optional[str] = None) -> Response:
        """Get child resources of an aggregation.

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: HTTP response from service, containing a list of textgrid metadata entries
        """
        response = self._requests.get(self._url + '/navigation/agg/' +
                                textgrid_uri, params={'sid': sid},
                                timeout=self._config.http_timeout)
        return self._handle_response(response)

    # pylint: disable-msg=too-many-arguments,too-many-locals
    def search(self,
               query: Optional[str] = '*',
               sid: Optional[str] = None,
               target: Optional[str] = None,
               order: Optional[str] = None,
               start: Optional[int] = None,
               limit: Optional[int] = None,
               kwic_width: Optional[int] = None,
               word_distance: Optional[int] = None,
               path: Optional[bool] = None,
               all_projects: Optional[bool] = None,
               sandbox: Optional[bool] = None,
               filters: Optional[List[str]] = None,
               facet:  Optional[List[str]] = None,
               facet_limit: Optional[int] = None) -> Response:
        """Run fulltext queries or filters on TextGrid metadata and fulltext objects.

        Please note: as the defaults of this function are mostly set to None, the defaults from
            the service are used, and also noted in this docstring. see:
            http://textgridlab.org/doc/services/submodules/tg-search/docs/api/search.html

        Args:
            query (str, optional): Lucene search string. Defaults to '*'.
            sid (str, optional): TextGrid SessionID from tgauth. Defaults to None.
            target (str, optional): where to do fulltext-searches: one of 'structure',
                                    'metadata' and 'both'. Defaults to 'both'.
            order (str, optional): key-value ascending (asc) or descending (desc) and metadata-field
                                   like asc:title or desc:author. Defaults to 'relevance'.
            start (int, optional): result number to start with.. Defaults to 0.
            limit (int, optional): number of entries to return.. Defaults to 20.
            kwic_width (int, optional): number of chars before and after a kwic match.
                                        Defaults to 40.
            word_distance (int, optional): max distance beetween two words in fulltext query.
                                           ignored if set to a number < 0, then for a hit all words
                                           must be contained in one document. Defaults to -1.
            path (boo], optional): path of found result(work->edition->aggregations) should be
                                   applied to hit. Defaults to false.
            all_projects (bool, optional): all Projects should be searched for public data,
                                           warning: this query may be slow, if many results found.
                                           Defaults to false.
            sandbox (bool, optional): show sandboxed (not yet finally published) data.
                                      Defaults to false.
            filters (List[str], optional): add filter on query results, e.g. for faceting.
                                           Defaults to None.
            facet (List[str], optional): get facets for query results. Defaults to None.
            facet_limit (int, optional): number of results to return for each facet. Defaults to 10.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: HTTP response from service - a list of textgrid metadata entries,
                      KWIC hits, paths and facets if requested
        """

        params = {
            'q': query,
            'sid': sid,
            'target': target,
            'order': order,
            'start': start,
            'limit': limit,
            'kwicWidth': kwic_width,
            'wordDistance': word_distance,
            'path': path,
            'allProjects': all_projects,
            'sandbox': sandbox,
            'filter': filters,
            'facet': facet,
            'facetLimit': facet_limit
        }
        response = self._requests.get(self._url + '/search', params=params,
                                timeout=self._config.http_timeout)
        return self._handle_response(response)
    # pylint: enable-msg=too-many-arguments,too-many-locals

    def edition_work_metadata_for(self, textgrid_uri: str, sid: Optional[str] = None) -> Response:
        """Find parent edition for an object and the edition and work metadata

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: HTTP response from service - edition and work metadata for given object
                      from first matching parent edition
        """
        response = self._requests.get(
            self._url + '/info/' + textgrid_uri +
            '/editionWorkMeta', params={'sid': sid},
            timeout=self._config.http_timeout)
        return self._handle_response(response)

    def children(self, textgrid_uri: str, sid: Optional[str] = None) -> Response:
        """List URIs for all children of this aggregation and its child aggregations

        Args:
            textgrid_uri (str): Textgrid URI of an aggregation
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            Response: HTTP response from service - URIs for children of this
                      aggregation and its child aggregations
        """
        response = self._requests.get(
            self._url + '/info/' + textgrid_uri + '/children', params={'sid': sid},
            timeout=self._config.http_timeout)
        return self._handle_response(response)

    @staticmethod
    def _handle_response(response: Response) -> Response:
        """Error handling for responses from tgsearch

        Args:
            response (Response): a response from tgsearch

        Raises:
            TextgridSearchException: if HTTP status code >= 400

        Returns:
            Response: the response
        """
        if not response.ok:
            message = '[Error] HTTP Code: ' + \
                str(response.status_code) + ' - ' + response.text[0:255]
            logger.warning(message)
            raise TextgridSearchException(message)
        return response


class TextgridSearch(TextgridSearchRequest):
    """Provide access to the TextGrid search service using a XML data binding """

    def __init__(self, config: TextgridConfig = TextgridConfig(), nonpublic: bool = False) -> None:
        super().__init__(config, nonpublic)
        # It’s recommended to either reuse the same parser/serializer instance
        # or reuse the context instance. see https://xsdata.readthedocs.io/en/latest/xml.html
        context = XmlContext()
        self._parser = XmlParser(context=context)

    def info(self, textgrid_uri: str, sid: Optional[str] = None) -> SearchResponse:
        """Retrieve metadata for a textgrid object specified by its
        textgrid-uri

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str]): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            SearchResponse: metadata for uri
        """
        response = super().info(textgrid_uri, sid)
        return self._parser.parse(BytesIO(response.content), SearchResponse)

    def list_project_root(self, project_id: str, sid: Optional[str] = None) -> SearchResponse:
        """Get objects belonging to a project, filtered by objects that are in
        an aggregation in the same project.

        Args:
            project_id (str): the ID of the project to list
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            SearchResponse: A list of textgrid metadata entries
        """
        response = super().list_project_root(project_id, sid)
        return self._parser.parse(BytesIO(response.content), SearchResponse)

    def list_aggregation(self, textgrid_uri: str, sid: Optional[str] = None) -> SearchResponse:
        """Get child resources of an aggregation.

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            SearchResponse: A list of textgrid metadata entries
        """
        response = super().list_aggregation(textgrid_uri, sid)
        return self._parser.parse(BytesIO(response.content), SearchResponse)

    # pylint: disable-msg=too-many-arguments,too-many-locals

    def search(self,
               query: Optional[str] = '*',
               sid: Optional[str] = None,
               target: Optional[str] = None,
               order: Optional[str] = None,
               start: Optional[int] = None,
               limit: Optional[int] = None,
               kwic_width: Optional[int] = None,
               word_distance: Optional[int] = None,
               path: Optional[bool] = None,
               all_projects: Optional[bool] = None,
               sandbox: Optional[bool] = None,
               filters: Optional[List[str]] = None,
               facet:  Optional[List[str]] = None,
               facet_limit: Optional[int] = None) -> SearchResponse:
        """Run fulltext queries or filters on TextGrid metadata and fulltext objects.

        Please note: as the defaults of this function are mostly set to None, the defaults from
            the service are used, and also noted in this docstring. see:
            http://textgridlab.org/doc/services/submodules/tg-search/docs/api/search.html

        Args:
            query (str, optional): Lucene search string. Defaults to '*'.
            sid (str, optional): TextGrid SessionID from tgauth. Defaults to None.
            target (str, optional): where to do fulltext-searches: one of 'structure',
                                    'metadata' and 'both'. Defaults to 'both'.
            order (str, optional): key-value ascending (asc) or descending (desc) and metadata-field
                                   like asc:title or desc:author. Defaults to 'relevance'.
            start (int, optional): result number to start with.. Defaults to 0.
            limit (int, optional): number of entries to return.. Defaults to 20.
            kwic_width (int, optional): number of chars before and after a kwic match.
                                        Defaults to 40.
            word_distance (int, optional): max distance beetween two words in fulltext query.
                                           ignored if set to a number < 0, then for a hit all words
                                           must be contained in one document. Defaults to -1.
            path (boo], optional): path of found result(work->edition->aggregations) should be
                                   applied to hit. Defaults to false.
            all_projects (bool, optional): all Projects should be searched for public data,
                                           warning: this query may be slow, if many results found.
                                           Defaults to false.
            sandbox (bool, optional): show sandboxed (not yet finally published) data.
                                      Defaults to false.
            filters (List[str], optional): add filter on query results, e.g. for faceting.
                                           Defaults to None.
            facet (List[str], optional): get facets for query results. Defaults to None.
            facet_limit (int, optional): number of results to return for each facet. Defaults to 10.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            SearchResponse: a list of textgrid metadata entries,
                            KWIC hits, paths and facets if requested
        """

        response = super().search(query=query,
                                  sid=sid,
                                  target=target,
                                  order=order,
                                  start=start,
                                  limit=limit,
                                  kwic_width=kwic_width,
                                  word_distance=word_distance,
                                  path=path,
                                  all_projects=all_projects,
                                  sandbox=sandbox,
                                  filters=filters,
                                  facet=facet,
                                  facet_limit=facet_limit)
        return self._parser.parse(BytesIO(response.content), SearchResponse)
    # pylint: enable-msg=too-many-arguments,too-many-locals

    def edition_work_metadata_for(self, textgrid_uri: str,
                                  sid: Optional[str] = None) -> SearchResponse:
        """Find parent edition for an object and the edition and work metadata

        Args:
            textgrid_uri (str): Textgrid URI
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            SearchResponse: Edition and work metadata for given object
                            from first matching parent edition
        """
        response = super().edition_work_metadata_for(textgrid_uri, sid)
        return self._parser.parse(BytesIO(response.content), SearchResponse)

    def children(self, textgrid_uri: str, sid: Optional[str] = None) -> TextgridUris:
        """List URIs for all children of this aggregation and its child aggregations

        Args:
            textgrid_uri (str): Textgrid URI of an aggregation
            sid (Optional[str], optional): Session ID. Defaults to None.

        Raises:
            TextgridSearchException: if HTTP status code >= 400   (# noqa: DAR402)

        Returns:
            TextgridUris: URIs for children of this aggregation and its child aggregations
        """
        response = super().children(textgrid_uri, sid)
        return self._parser.parse(BytesIO(response.content), TextgridUris)


class TextgridSearchException(Exception):
    """Exception communicating with tgsearch"""



# SPDX-FileCopyrightText: 2022 Georg-August-Universität Göttingen
#
# SPDX-License-Identifier: CC0-1.0

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union
from xsdata.models.datatype import XmlDate, XmlDateTime, XmlPeriod
from tgclients.databinding.rdf import Rdf
from tgclients.databinding.textgrid_metadata_agent_2010 import (
    AgentType,
    PersonType,
)
from tgclients.databinding.textgrid_metadata_script_2010 import FormOfNotationType

__NAMESPACE__ = "http://textgrid.info/namespaces/metadata/core/2010"


@dataclass
class AuthorityType:
    """
    references to authority files like thesauri etc.
    """
    class Meta:
        name = "authorityType"

    id: List["AuthorityType.Id"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )

    @dataclass
    class Id:
        """
        Attributes
            value:
            type: The type of the ID in a non-textgrid context
        """
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        type: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            }
        )


@dataclass
class DateType:
    """
    same function as our old sub-schema.
    """
    class Meta:
        name = "dateType"

    value: str = field(
        default="",
        metadata={
            "required": True,
        }
    )
    date: Optional[Union[XmlDate, XmlPeriod]] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    not_before: Optional[Union[XmlDate, XmlPeriod]] = field(
        default=None,
        metadata={
            "name": "notBefore",
            "type": "Attribute",
        }
    )
    not_after: Optional[Union[XmlDate, XmlPeriod]] = field(
        default=None,
        metadata={
            "name": "notAfter",
            "type": "Attribute",
        }
    )


class GeneratedTypeAvailability(Enum):
    """
    possible values.
    """
    DEFAULT = "default"
    PUBLIC = "public"
    STABLE = "stable"


@dataclass
class IdentifierType:
    """An unambiguous reference to the resource within a given context.

    Same
    as: http://purl.org/dc/terms/identifier

    Attributes
        value:
        type: The type of the URI in a non-TextGrid context (e.g., a
            ISBN or ISSN, a shelfmark, a registration or inventary
            number)
    """
    class Meta:
        name = "identifierType"

    value: str = field(
        default="",
        metadata={
            "required": True,
        }
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        }
    )


class PidPidType(Enum):
    HANDLE = "handle"
    URN = "urn"
    DOI = "doi"
    ARK = "ark"
    PURL = "purl"
    OTHER = "other"


class WorkTypeGenre(Enum):
    """The genre of the work. Same as:

    http://purl.org/dc/terms/type - Hier natürlich englische Bezeichnungen,
    wenn wir mal die Liste finalisiert haben. Siehe
    http://www.textgrid.de/intern/wiki1/wiki/Genres.html
    """
    DRAMA = "drama"
    PROSE = "prose"
    VERSE = "verse"
    REFERENCE_WORK = "reference work"
    NON_FICTION = "non-fiction"
    NON_TEXT = "non-text"
    OTHER = "other"


@dataclass
class BibliographicCitationType:
    """
    Attributes
        author: A person or organization chiefly responsible for the
            intellectual or artistic content of the source, usually
            printed text. Same as:
            http://id.loc.gov/vocabulary/relators/aut.html. For a
            detailed description of the author the "agent" element in
            the "workType" has to be used.
        editor: A person or organization who prepares for publication a
            work not primarily his/her own, such as by elucidating text,
            adding introductory or other critical matter, or technically
            directing an editorial staff. Same as:
            http://id.loc.gov/vocabulary/relators/edt.htmln
        edition_title: Name of the source (e.g., the titel of a book or
            a journal).
        place_of_publication: Place where the source was published.
        publisher: A person or organization that makes the source
            available to the public. Same as:
            http://id.loc.gov/vocabulary/relators/pbl.html
        date_of_publication: TG dateType.
        edition_no: Statement of the edition of the source - usually a
            phrase with or without numbers (e.g., first edition, 4th
            ed., etc.)
        series: Titel of a series in which the source was issued.
        volume: Volume designation - is usually expressed as a number
            but could be roman numerals or non-numeric (e.g., 124, VI,
            etc.)
        issue: Designation of the issue of a journal the source was
            published. While usually numeric, it could be non-numeric
            (e.g. Spring)
        spage: Designates the page the source starts in a volume or
            issue. Pages are not always numeric.
        epage: Designates the page the source ends in a volume or issue.
            Page are not always numeric.
        bib_identifier:
    """
    class Meta:
        name = "bibliographicCitationType"

    author: List[PersonType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    editor: List[PersonType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    edition_title: List[str] = field(
        default_factory=list,
        metadata={
            "name": "editionTitle",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    place_of_publication: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "name": "placeOfPublication",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    publisher: List[PersonType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    date_of_publication: Optional[DateType] = field(
        default=None,
        metadata={
            "name": "dateOfPublication",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    edition_no: Optional[str] = field(
        default=None,
        metadata={
            "name": "editionNo",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    series: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    volume: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    issue: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    spage: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    epage: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    bib_identifier: Optional[IdentifierType] = field(
        default=None,
        metadata={
            "name": "bibIdentifier",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class CollectionType:
    """
    TODO.

    Attributes
        collector: Person or corporate body creating a collection. Same
            as: http://id.loc.gov/vocabulary/relators/col.html
        abstract: A summary of the collection. Same as:
            http://purl.org/dc/terms/abstract
        collection_description: URL of a description of this collection
            (e.g., project description on the website of a funding
            agency) FIXME is this still the Same as:
            http://purl.org/dc/terms/description
        spatial: Spatial characteristics of the collection. Same as:
            http://purl.org/dc/terms/spatial
        temporal: Temporal characteristics of the collection. Same as:
            http://purl.org/dc/terms/temporal
        subject: The topic of the collection. Same as:
            http://purl.org/dc/terms/subject.
    """
    class Meta:
        name = "collectionType"

    collector: List[PersonType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    abstract: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    collection_description: List[str] = field(
        default_factory=list,
        metadata={
            "name": "collectionDescription",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    spatial: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    temporal: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    subject: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class GeneratedType:
    """The generatedType is cerated by the middleware (TG-crud), and is
    delivered with every T-crud call, that delivers back the metadata
    ObjectType.

    With every #UPDATE or #UPDATEMETADATA call this generatedType must
    also be sent in the request parameter's objectType, to be able to
    handle concurrent modification of TextGrid objects.

    Attributes
        created: Date of creation of the resource. Same as:
            http://purl.org/dc/terms/created. TODO Datentyp
            spezifizieren
        last_modified: Date on which the resource was changed. Same as:
            http://purl.org/dc/terms/modified. TODO Datentyp
            spezifizieren
        issued: Date of formal issuance (e.g., publication) of the
            resource. Same as: http://purl.org/dc/terms/issued. TODO
            Datentyp (oder xs:date?) spezifizieren
        textgrid_uri: Subproperty of identifier. An unambiguous
            reference to the resource within the textgrid context.
            Subproperty of identifier:
            http://purl.org/dc/elements/1.1/identifier. Hier in jedem
            Falle die TextGrid-URI (NOID). FIXME für die Fälle (1)
            veröffentlicht (PID – oder kommt der in
            ../provided/identifier?), (2) externe Datenquelle (oder auch
            der nicht hier?) noch weitere Identifier hier? Sonst
            Kardinalität auf 1..1 setzen.
        revision: revision number
        pid: persistent identifier
        extent: The size of the resource in bytes. Same as:
            http://purl.org/dc/terms/extent.
        fixity: "Information used to verify whether an object has been
            altered in an undocumented or unauthorized way." - stolen
            from the PREMIS Object Entity sub-schema, see
            http://www.loc.gov/standards/premis/v1/Object-v1-1.xsd. For
            detailed documentation see the PREMIS Data Dictionary
            (http://www.loc.gov/standards/premis/v2/premis-dd-2-0.pdf)
            section 1.5.2
        data_contributor: Use for a person that submits data in
            textgrid. Same as:
            http://lcweb2.loc.gov/diglib/loc.terms/relators/DTC.
        project: Project name and id.
        warning: Error report provided by the middleware when an
            incorrect textgrid object has been stored.
        permissions: Same as accessRights ("Information about who can
            access the resource or an indication of its security status"
            http://purl.org/dc/terms/accessRights) in textgrid? In
            TextGrid, this is dependent on the role of the person that
            accesses the resource.
        availability: information whether the TG object is subject of
            the (role based) rights management or isPublic or isStable
        any_element: Reserved for future additions
    """
    class Meta:
        name = "generatedType"

    created: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    last_modified: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "lastModified",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    issued: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    textgrid_uri: Optional["GeneratedType.TextgridUri"] = field(
        default=None,
        metadata={
            "name": "textgridUri",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    revision: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    pid: List["GeneratedType.Pid"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    extent: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    fixity: List["GeneratedType.Fixity"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    data_contributor: Optional[str] = field(
        default=None,
        metadata={
            "name": "dataContributor",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    project: Optional["GeneratedType.Project"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    warning: List["GeneratedType.Warning"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    permissions: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    availability: Optional[GeneratedTypeAvailability] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    any_element: List[object] = field(
        default_factory=list,
        metadata={
            "type": "Wildcard",
            "namespace": "##any",
        }
    )

    @dataclass
    class TextgridUri:
        """
        Attributes
            value:
            ext_ref: The TG metadata refer to an external object
        """
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        ext_ref: Optional[str] = field(
            default=None,
            metadata={
                "name": "extRef",
                "type": "Attribute",
            }
        )

    @dataclass
    class Pid:
        """
        Attributes
            value:
            pid_type: PID type
        """
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        pid_type: PidPidType = field(
            default=PidPidType.HANDLE,
            metadata={
                "name": "pidType",
                "type": "Attribute",
            }
        )

    @dataclass
    class Fixity:
        """
        Attributes
            message_digest_algorithm: The specific algorithm used to
                construct the message digest for the digital object,
                e.g. MD5, SHA-1..n etc.
            message_digest: The output of the message digest algorithm,
                i.e. the checksum
            message_digest_originator: The agent that created the
                original message digest that is compared in a fixity
                check. In TextGrid: "TG-crud" or Service-endpoint?
        """
        message_digest_algorithm: Optional[str] = field(
            default=None,
            metadata={
                "name": "messageDigestAlgorithm",
                "type": "Element",
                "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
                "required": True,
            }
        )
        message_digest: Optional[str] = field(
            default=None,
            metadata={
                "name": "messageDigest",
                "type": "Element",
                "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
                "required": True,
            }
        )
        message_digest_originator: Optional[str] = field(
            default=None,
            metadata={
                "name": "messageDigestOriginator",
                "type": "Element",
                "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            }
        )

    @dataclass
    class Project:
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        id: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            }
        )

    @dataclass
    class Warning:
        """
        Attributes
            value:
            uri: The URI the warning occured.
        """
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        uri: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            }
        )


@dataclass
class ItemType:
    """
    Attributes
        rights_holder: Person or organization with copyright for an
            item. Same as: http://purl.org/dc/terms/rightsHolder FIXME
            dcterms-Problem wieder: Damit ist möglicherweise auch die
            Entscheidung, als Datentyp für die diversen Personenfelder
            wie rightsHolder etc. dcterms:irgendwas zu nehmen,
            hinfällig, da damit wohl keine Möglichkeit zur Spezifikation
            einer ID zusätzlich zum Inhaltsstring (Namen) gegeben ist.
    """
    class Meta:
        name = "itemType"

    rights_holder: List[PersonType] = field(
        default_factory=list,
        metadata={
            "name": "rightsHolder",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class ObjectCitationType:
    """
    Attributes
        object_title: Name of the source (e.g., the name of a painting,
            sculpture, furniture, etc. ).
        object_contributor: A person or organization responsible for the
            intellectual or artistic content of a work. Same as:
            http://purl.org/dc/terms/contributor.
        object_date: Date of an event in the lifecycle of the resource.
            Same as: http://purl.org/dc/terms/date. For a detailed
            description of the date of creation the "dateOfCreation" and
            "timeOfCreation" elements in the "workType" have to be used.
        object_identifier:
    """
    class Meta:
        name = "objectCitationType"

    object_title: List[str] = field(
        default_factory=list,
        metadata={
            "name": "objectTitle",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    object_contributor: List[AgentType] = field(
        default_factory=list,
        metadata={
            "name": "objectContributor",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    object_date: Optional[DateType] = field(
        default=None,
        metadata={
            "name": "objectDate",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    object_identifier: Optional[IdentifierType] = field(
        default=None,
        metadata={
            "name": "objectIdentifier",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class ProvidedType:
    """
    Attributes
        title: A name given to the resource. Same as:
            http://purl.org/dc/terms/title; FIXME Uniform-Title
            allerdings ist 1..1 :-(
        identifier:
        format: File format of a textgrid object. Same as:
            http://purl.org/dc/terms/format. TODO Specification of type.
        notes: Anything that doesn't fit into the other fields.
    """
    class Meta:
        name = "providedType"

    title: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    identifier: List[IdentifierType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    format: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    notes: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class RelationType:
    """
    Attributes
        is_derived_from: Link to an object the described textgrid object
            is derived from.
        is_alternative_format_of: Link to an object having the same
            content as the described textgrid object but in another
            format.
        has_adaptor: Link to an object which is an adaptor of the
            described textgrid object.
        has_schema: Link to an object which is a schema of the described
            textgrid object.
        rdf:
    """
    class Meta:
        name = "relationType"

    is_derived_from: Optional[str] = field(
        default=None,
        metadata={
            "name": "isDerivedFrom",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    is_alternative_format_of: Optional[str] = field(
        default=None,
        metadata={
            "name": "isAlternativeFormatOf",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    has_adaptor: Optional[str] = field(
        default=None,
        metadata={
            "name": "hasAdaptor",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    has_schema: Optional[str] = field(
        default=None,
        metadata={
            "name": "hasSchema",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    rdf: Optional[Rdf] = field(
        default=None,
        metadata={
            "name": "RDF",
            "type": "Element",
            "namespace": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        }
    )


@dataclass
class WorkType:
    """
    FIXME Grobklassifikation.

    Attributes
        agent:
        abstract: A summary of the work. Same as:
            http://purl.org/dc/terms/abstract
        date_of_creation: unser (TextGrid) Datumsmodell FIXME: "created"
            gibts schon in generated
        spatial: Spatial characteristics of the work. Same as:
            http://purl.org/dc/terms/spatial
        temporal: Temporal characteristics of the work. Same as:
            http://purl.org/dc/terms/temporal
        subject: The topic of the work. Same as:
            http://purl.org/dc/terms/subject.
        genre: Grobklassifikation
        type: The nature or genre of the work. Same as:
            http://purl.org/dc/terms/type FIXME Feinklassifikation,
            editierbare Auswahlliste
    """
    class Meta:
        name = "workType"

    agent: List[AgentType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    abstract: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    date_of_creation: Optional[DateType] = field(
        default=None,
        metadata={
            "name": "dateOfCreation",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    spatial: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    temporal: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    subject: List[AuthorityType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    genre: List[WorkTypeGenre] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    type: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class GenericType:
    """
    Attributes
        provided: Metadata provided by the client.
        generated: Metadata generated by the middleware components.
    """
    class Meta:
        name = "genericType"

    provided: Optional[ProvidedType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )
    generated: Optional[GeneratedType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class SourceType:
    """Relation between the textgrid object and the description of the source
    of this object.

    Same as: http://purl.org/dc/terms/source
    """
    class Meta:
        name = "sourceType"

    bibliographic_citation: Optional[BibliographicCitationType] = field(
        default=None,
        metadata={
            "name": "bibliographicCitation",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    object_citation: Optional[ObjectCitationType] = field(
        default=None,
        metadata={
            "name": "objectCitation",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )


@dataclass
class EditionType:
    """
    Attributes
        is_edition_of: Manifestation of which work? Value must be the
            TextGrid URI of a TextGrid Work object. Field is mandatory
            on publication. Same as:
            http://rdvocab.info/RDARelationshipsWEMI/manifestationOfWork
        agent:
        source:
        form_of_notation:
        language:
        license: A legal document giving official permission to do
            something with the resource. Same as:
            http://purl.org/dc/terms/license.
    """
    class Meta:
        name = "editionType"

    is_edition_of: Optional[str] = field(
        default=None,
        metadata={
            "name": "isEditionOf",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    agent: List[AgentType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "min_occurs": 1,
        }
    )
    source: List[SourceType] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    form_of_notation: List[FormOfNotationType] = field(
        default_factory=list,
        metadata={
            "name": "formOfNotation",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )
    language: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "pattern": r"[a-z][a-z][a-z]",
        }
    )
    license: Optional["EditionType.License"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
        }
    )

    @dataclass
    class License:
        """
        Attributes
            value:
            license_uri: The context in which a license originates.
                Recommendation is to use Creative Commons (see:
                http://creativecommons.org/licenses/)
        """
        value: str = field(
            default="",
            metadata={
                "required": True,
            }
        )
        license_uri: Optional[str] = field(
            default=None,
            metadata={
                "name": "licenseUri",
                "type": "Attribute",
            }
        )


@dataclass
class Object:
    """
    The root element should also be used in CRUD arguments and returned by TG-
    search whenever TG-search returns complete metadata records.

    Attributes
        generic: Metadata available in all kinds of TextGrid objects.
        item: Item specific metadata.
        edition: Metadata specific for editions. - An embodiment of a
            work. Broader than:
            http://rdvocab.info/uri/schema/FRBRentitiesRDA/Manifestation
        work: Metadata specific for works. - A distinct intellectual or
            artistic creation. Broader than:
            http://rdvocab.info/uri/schema/FRBRentitiesRDA/Work
        collection: FIXME - je nach Sammlungs-Konzept und dessen
            Umsetzung - public-Bereich
        custom: The custom field may contain any additional metadata
            records for specific projects or specific applications.
        relations: FIXME wir haben ein paar spezifische Relationen,
            wollen aber auch beliebige Beziehungen durch die Anwender
            erlauben. Die Optionen hier sind vermutlich: (a) wir
            erlauben beliebiges RDF/XML und werten bestimmte
            bestandteile aus. (b) wir erlauben "unsere" relationen +
            beliebiges RDF/XML, also eine Substruktur
    """
    class Meta:
        name = "object"
        namespace = "http://textgrid.info/namespaces/metadata/core/2010"

    generic: Optional[GenericType] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    item: Optional[ItemType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    edition: Optional[EditionType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    work: Optional[WorkType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    collection: Optional[CollectionType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    custom: Optional["Object.Custom"] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    relations: Optional[RelationType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )

    @dataclass
    class Custom:
        other_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##other",
            }
        )


@dataclass
class MetadataContainerType:
    """
    Container type that contains exactly one metadata record, complete with
    it's root element (tns:object)
    """
    class Meta:
        name = "metadataContainerType"

    object_value: Optional[Object] = field(
        default=None,
        metadata={
            "name": "object",
            "type": "Element",
            "namespace": "http://textgrid.info/namespaces/metadata/core/2010",
            "required": True,
        }
    )


@dataclass
class TgObjectMetadata(MetadataContainerType):
    """
    Element for the container type.
    """
    class Meta:
        name = "tgObjectMetadata"
        namespace = "http://textgrid.info/namespaces/metadata/core/2010"
